@@ -1,7 +1,22 @@
-.onAttach <- function(libname, pkgname){ ## nocov start
+orig.onAttach <- function(libname, pkgname){
+    ## nocov start
     ## Setup RxODE.prefer.tbl
     nlmixrSetupMemoize()
     options(keep.source=TRUE)
+    ## nocov end
+}
+
+##' Convert data to RxODE format (depreciated)
+##'
+##' @param data Data to "convert"
+##'
+##' @return Exact same data as was input
+##'
+##' @keywords internal
+##' @export
+nmDataConvert <- function(data){
+    warning("nmDataConvert is depreciated and no longer needed.");
+    data
 }
 
 nlmixrSetupMemoize <- function(){
@@ -54,7 +69,7 @@ nlmixrForget <- function(){
 ##' @importFrom Rcpp evalCpp
 ##' @importFrom dparser dparse
 ##' @importFrom vpc vpc
-##' @importFrom ggplot2 ggplot aes geom_point facet_wrap geom_line geom_abline xlab geom_smooth
+##' @importFrom ggplot2 ggplot aes geom_point facet_wrap geom_line geom_abline xlab geom_smooth aes_string
 ##' @importFrom RcppArmadillo armadillo_version
 ##' @useDynLib nlmixr, .registration=TRUE
 
@@ -85,6 +100,8 @@ armaVersion <- function(){
     nlmixrLogo(str="RcppArmadiilo", RcppArmadillo::armadillo_version())
 }
 
+.nlmixrTime <- NULL
+
 ##' nlmixr fits population PK and PKPD non-linear mixed effects models.
 ##'
 ##' nlmixr is an R package for fitting population pharmacokinetic (PK)
@@ -98,17 +115,24 @@ armaVersion <- function(){
 ##' @param object Fitted object or function specifying the model.
 ##' @inheritParams nlmixr_fit
 ##' @param ... Other parameters
+##' @param save Boolean to save a nlmixr object in a rds file in the
+##'     working directory.  If \code{NULL}, uses option "nlmixr.save"
 ##' @return Either a nlmixr model or a nlmixr fit object
 ##' @author Matthew L. Fidler, Rik Schoemaker
 ##' @export
 nlmixr <- function(object, data, est=NULL, control=list(),
-                   table=tableControl(), ...){
+                   table=tableControl(), ...,save=NULL){
+    assignInMyNamespace(".nlmixrTime",proc.time());
+    force(est)
+    ## verbose?
+    ## https://tidymodels.github.io/model-implementation-principles/general-conventions.html
     UseMethod("nlmixr")
 }
 
 ##' @rdname nlmixr
 ##' @export
-nlmixr.function <- function(object, data, est=NULL, control=list(), table=tableControl(), ...){
+nlmixr.function <- function(object, data, est=NULL, control=list(), table=tableControl(), ...,
+                            save=NULL){
     .args <- as.list(match.call(expand.dots=TRUE))[-1]
     .uif <- nlmixrUI(object);
     class(.uif) <- "list";
@@ -120,21 +144,49 @@ nlmixr.function <- function(object, data, est=NULL, control=list(), table=tableC
         .uif$nmodel$data.name <- deparse(substitute(data))
         class(.uif) <- "nlmixrUI"
         .args$data <- data;
+        .args$est  <- est
         .args <- c(list(uif=.uif), .args[-1]);
+        if (is.null(est)){
+            stop("Need to supply an estimation routine with est=.");
+        }
         return(do.call(nlmixr_fit, .args))
     }
 }
 
+##'@rdname nlmixr
+##'@export
+nlmixr.nlmixrFitCore <- function(object, data, est=NULL, control=list(), table=tableControl(), ...,
+                                 save=NULL){
+    .uif <- .getUif(object);
+    if (missing(data)){
+        data <- getData(object);
+    }
+    .args <- as.list(match.call(expand.dots=TRUE))[-1]
+    .args$data <- data;
+    .args$est <- est;
+    .args <- c(list(uif=.uif), .args[-1]);
+    return(do.call(nlmixr_fit, .args))
+}
+
 ##' @rdname nlmixr
 ##' @export
-nlmixr.nlmixrUI <- function(object, data, est=NULL, control=list(), ...){
+nlmixr.nlmixrUI <- function(object, data, est=NULL, control=list(), ...,
+                            save=NULL){
+    .args <- as.list(match.call(expand.dots=TRUE))[-1]
     .uif <- object
     if (missing(data) && missing(est)){
         return(.uif)
     } else {
         .args <- c(list(uif=.uif), .args[-1]);
-        .uif$nmodel$data.name <- deparse(substitute(data))
-        .args$data <- data;
+        if (missing(data) && !is.null(.getPipedData())){
+            data <- .getPipedData();
+            .args$data <- data;
+            .args$est <- est;
+        } else {
+            .uif$nmodel$data.name <- deparse(substitute(data))
+            .args$data <- data;
+            .args$est <- est;
+        }
         return(do.call(nlmixr_fit, .args));
     }
 }
@@ -143,16 +195,18 @@ nlmixr.nlmixrUI <- function(object, data, est=NULL, control=list(), ...){
 ##'
 ##' @param data is the name of the data to convert.  Can be a csv file
 ##'     as well.
+##' @param model This is the RxODE model to use to translate against
+##'     when parsing the data.
 ##' @return Appropriately formatted data
 ##' @author Matthew L. Fidler
 ##' @keywords internal
 ##' @export
-nlmixrData <- function(data){
+nlmixrData <- function(data, model=NULL){
     UseMethod("nlmixrData");
 }
 ##' @export
 ##' @rdname nlmixrData
-nlmixrData.character <- function(data){
+nlmixrData.character <- function(data, model=NULL){
     if (!file.exists(data)){
         stop(sprintf("%s does not exist.", data))
     }
@@ -164,91 +218,25 @@ nlmixrData.character <- function(data){
 }
 ##' @export
 ##' @rdname nlmixrData
-nlmixrData.default <- function(data){
-    dat <- as.data.frame(data);
-    nm1 <- toupper(names(dat));
-    for (n in c("ID", "EVID", "TIME", "DV", "AMT")){
-        w <- which(nm1 == n)
-        if (length(w) == 1L){
-            names(dat)[w] <- n;
-        } else if (length(w) == 0L){
-            stop(sprintf("Need '%s' data item in dataset.", n))
-        } else {
-            stop(sprintf("Multiple '%s' columns in dataset.", n))
-        }
-    }
-    if (is(dat$ID, "factor")){
-        dat$ID <- paste(dat$ID);
-    }
-    idSort <- .Call(`_nlmixr_chkSortIDTime`, as.integer(dat$ID), as.double(dat$TIME), as.integer(dat$EVID));
-    if (idSort == 3L){
-        dat <- dat[.Call(`_nlmixr_allDose`, as.integer(dat$EVID), as.integer(dat$ID)), ]
-        ## warning("NONMEM-style data converted to nlmixr/RxODE-style data.");
-        return(nlmixrData.default(nmDataConvert(dat)));
-    }
-    backSort <- c()
-    backSort2 <- c();
-    if (is(dat$ID, "character")){
-        ## Character need to sort first
-        lvl <- unique(dat$ID);
-        lab <- paste(lvl)
-        dat$ID <- factor(dat$ID, levels=lvl, labels=lab);
-        dat <- dat[.Call(`_nlmixr_allDose`, as.integer(dat$EVID), as.integer(dat$ID)), ];
-        ## Dropped some IDs?  Need to make sure they are still sequential
-        dat$ID <- factor(dat$ID, levels=lvl, labels=lab);
-        ## Get new factor level
-        dat$ID <- paste(dat$ID);
-        lvl <- unique(dat$ID);
-        lab <- paste(lvl)
-        dat$ID <- factor(dat$ID, levels=lvl, labels=lab);
-        backSort <- levels(dat$ID);
-        backSort2 <- seq_along(backSort)
-        dat$ID <- as.integer(dat$ID);
+nlmixrData.default <- function(data, model=NULL){
+    if (!is.null(model)){
+        dat  <- RxODE::etTrans(data, model,addCmt=TRUE,dropUnits=TRUE,allTimeVar=TRUE)
     } else {
-        if (idSort == 2L){
-            dat <- dat[.Call(`_nlmixr_allDose`, as.integer(dat$EVID), as.integer(dat$ID)), ];
-            lvl <- unique(dat$ID);
-            lab <- paste(lvl)
-            dat$ID <- factor(dat$ID, levels=lvl, labels=lab);
-            backSort <- levels(dat$ID);
-            backSort2 <- seq_along(backSort)
-            dat$ID <- as.integer(dat$ID);
-        } else if (idSort == 0L){
-            warning("Sorting by ID, TIME; Output fit may not be in the same order as input dataset.")
-            dat <- dat[order(dat$ID, dat$TIME), ];
-            dat <- dat[.Call(`_nlmixr_allDose`, as.integer(dat$EVID), as.integer(dat$ID)), ]
-            lvl <- unique(dat$ID);
-            lab <- paste(lvl)
-            dat$ID <- factor(dat$ID, levels=lvl, labels=lab);
-            backSort <- levels(dat$ID);
-            backSort2 <- seq_along(backSort)
-            dat$ID <- as.integer(dat$ID);
-        }
+        dat <- as.data.frame(data);
     }
-    attr(dat, "backSort") <- backSort;
-    attr(dat, "backSort2") <- backSort2;
     return(dat);
 }
-
-##' Fit a nlmixr model
-##'
-##' @param uif Parsed nlmixr model (by \code{nlmixr(mod.fn)}).
-##' @param data Dataset to estimate.  Needs to be RxODE compatible in
-##'     EVIDs.
-##' @param est Estimation method
-##' @param control Estimation control options.  They could be
-##'     \code{\link[nlme]{nlmeControl}}, \code{\link{saemControl}} or
-##'     \code{\link{foceiControl}}
-##' @param ... Parameters passed to estimation method.
-##' @param sum.prod Take the RxODE model and use more precise
-##'     products/sums.  Increases solving accuracy and solving time.
-##' @param table A list controlling the table options (i.e. CWRES,
-##'     NPDE etc).  See \code{\link{tableControl}}.
-##' @return nlmixr fit object
-##' @author Matthew L. Fidler
-##' @export
-nlmixr_fit <- function(uif, data, est=NULL, control=list(), ...,
-                       sum.prod=FALSE, table=tableControl()){
+nlmixr_fit0 <- function(uif, data, est=NULL, control=list(), ...,
+                        sum.prod=FALSE, table=tableControl()){
+    .clearPipedData();
+    .tmp <- deparse(body(uif$theta.pars))[-1];
+    .tmp <- .tmp[-length(.tmp)];
+    .origData <- data;
+    data <- RxODE::etTrans(data,paste(paste(.tmp,collapse="\n"),"\n",uif$rxode),TRUE,TRUE, TRUE);
+    .nTv  <- attr(class(data),".RxODE.lst")$nTv;
+    .lab  <- attr(class(data),".RxODE.lst")$idLvl;
+    .modelId <- digest::digest(list(sessionInfo()$otherPkgs$nlmixr$Version,
+                                    uif, data, est, control, sum.prod, table,...));
     .meta <- uif$meta
     .missingEst <- is.null(est);
     if (.missingEst & exists("est", envir=.meta)){
@@ -273,7 +261,11 @@ nlmixr_fit <- function(uif, data, est=NULL, control=list(), ...,
     }
     start.time <- Sys.time();
     if (!is(table, "tableControl")){
-        table <- do.call(tableControl, table);
+        if (is(table, "list")){
+            table <- do.call(tableControl, table);
+        } else {
+            table <- tableControl();
+        }
     }
     dat <- nlmixrData(data);
     up.covs <- toupper(uif$all.covs);
@@ -284,19 +276,20 @@ nlmixr_fit <- function(uif, data, est=NULL, control=list(), ...,
             names(dat)[w] = uif$all.covs[i];
         }
     }
-    backSort <- attr(dat, "backSort");
-    backSort2 <- attr(dat, "backSort2");
-    attr(dat, "backSort") <- NULL;
-    attr(dat, "backSort2") <- NULL;
-    uif$env$infusion <- .Call(`_nlmixr_chkSolvedInf`, as.double(dat$EVID), as.integer(!is.null(uif$nmodel$lin.solved)));
+    ## backSort <- attr(dat, "backSort");
+    ## backSort2 <- attr(dat, "backSort2");
+    ## attr(dat, "backSort") <- NULL;
+    ## attr(dat, "backSort2") <- NULL;
+    if (!is.null(uif$nmodel$lin.solved)==1L){
+        uif$env$infusion <- max(dat$EVID)>10000
+    }
     bad.focei <- "Problem calculating residuals, returning fit without residuals.";
     fix.dat <- function(x){
-        if (length(backSort) > 0){
-            cls <- class(x);
-            class(x) <- "data.frame";
-            x$ID <- factor(x$ID, backSort2, labels=backSort);
-            class(x) <- cls;
-        }
+        .cls <- class(x);
+        class(x) <- "data.frame";
+        attr(x$ID, "levels") <- .lab;
+        class(x$ID) <- "factor";
+        class(x) <- .cls;
         .uif <- x$uif;
         .thetas <- x$theta;
         for (.n in names(.thetas)){
@@ -309,6 +302,7 @@ nlmixr_fit <- function(uif, data, est=NULL, control=list(), ...,
             }
         }
         .env <- x$env
+        .env$origData <- .origData;
         .env$uif <- .uif;
         .predDf <- .uif$predDf;
         if (any(.predDf$cond != "") & any(names(x) == "CMT")){
@@ -350,7 +344,11 @@ nlmixr_fit <- function(uif, data, est=NULL, control=list(), ...,
         }
     }
     if (est == "saem"){
+        if (.nTv != 0) stop("SAEM does not support time-varying covariates (yet)");
         pt <- proc.time()
+        if (length(uif$noMuEtas) > 0){
+            stop(sprintf("Cannot run SAEM since some of the parameters are not mu-referenced (%s)", paste(uif$noMuEtas, collapse=", ")))
+        }
         args <- as.list(match.call(expand.dots=TRUE))[-1]
         default <- saemControl();
         if (any(names(args) == "mcmc")){
@@ -400,51 +398,94 @@ nlmixr_fit <- function(uif, data, est=NULL, control=list(), ...,
         } else {
             .logLik <- default$logLik
         }
+        if (any(names(control) == "nnodes.gq")){
+            .nnodes.gq <- control$nnodes.gq
+        } else {
+            .nnodes.gq <- default$nnodes.gq
+        }
+        if (any(names(control) == "nsd.gq")){
+            .nsd.gq <- control$nsd.gq
+        } else {
+            .nsd.gq <- default$nsd.gq
+        }
+        if (any(names(control) == "adjObf")){
+            .adjObf <- control$adjObf;
+        } else {
+            .adjObf <- default$adjObf;
+        }
         if (any(names(control) == "optExpression")){
             uif$env$optExpression <- control$optExpression
         } else {
             uif$env$optExpression <- default$optExpression
         }
-
+        if (any(names(control) == "covMethod")){
+            .addCov <- control$covMethod == "linFim";
+        } else {
+            .addCov <- default$covMethod == "linFim";
+        }
+        if (uif$saemErr!=""){
+            stop(paste0("For SAEM:\n",uif$saemErr))
+        }
         if (is.null(uif$nlme.fun.mu)){
             stop("SAEM requires all ETAS to be mu-referenced")
         }
+        .err  <- uif$ini$err
         .low <- uif$ini$lower;
-        .low <- .low[!is.na(.low)]
+        .low <- .low[!is.na(.low) & is.na(.err)]
         .up <- uif$ini$upper;
-        .up <- .up[!is.na(.up)]
+        .up <- .up[!is.na(.up) & is.na(.err)]
         if (any(.low != -Inf) | any(.up != Inf)){
             warning("Bounds are ignored in SAEM")
         }
-        if (any(paste(uif$ini$name[uif$ini$fix]) %in% unlist(uif$mu.ref))){
-            stop("Fixed thetas cannot be associated with an ETA in SAEM")
-        }
         uif$env$mcmc <- mcmc;
         uif$env$ODEopt <- ODEopt;
         uif$env$sum.prod <- sum.prod
         uif$env$covMethod <- covMethod
         .dist <- uif$saem.distribution
         model <- uif$saem.model
-        cfg <- configsaem(model=model, data=dat, inits=uif$saem.init,
-                          mcmc=mcmc, ODEopt=ODEopt, seed=seed, fixed=uif$saem.fixed,
+        inits = uif$saem.init
+        if (length(uif$saem.fixed)>0) {
+			nphi = attr(model$saem_mod, "nrhs")
+			m = cumsum(!is.na(matrix(inits$theta, byrow=T, ncol=nphi)))
+			fixid = match(uif$saem.fixed, t(matrix(m,ncol=nphi)))
+
+			names(inits$theta) = rep("", length(inits$theta))
+			names(inits$theta)[fixid] = "FIXED"
+        }
+        cfg <- configsaem(model=model, data=dat, inits=inits,
+                          mcmc=mcmc, ODEopt=ODEopt, seed=seed,
                           distribution=.dist, DEBUG=DEBUG);
-        if (print > 1){
+        if (is(print, "numeric")){
             cfg$print <- as.integer(print)
         }
         .fit <- model$saem_mod(cfg);
-        .ret <- as.focei.saemFit(.fit, uif, pt, data=dat, calcResid=calc.resid, obf=.logLik);
+        .ret <- try(as.focei.saemFit(.fit, uif, pt, data=dat, calcResid=calc.resid, obf=.logLik,
+                                     nnodes.gq=.nnodes.gq, nsd.gq=.nsd.gq, adjObf=.adjObf,
+                                     addCov=.addCov), silent=TRUE);
+        if (inherits(.ret, "try-error")){
+            warning("Error converting to nlmixr UI object, returning saem object");
+            return(.fit)
+        }
         if (inherits(.ret, "nlmixrFitData")){
             .ret <- fix.dat(.ret);
             .ret <- .addNpde(.ret);
         }
         if (inherits(.ret, "nlmixrFitCore")){
             .env <- .ret$env
+            .env$adjObj <- .adjObf;
+            .env$nnodes.gq <- .nnodes.gq;
+            .env$nsd.gq <- .nsd.gq
             assign("startTime", start.time, .env);
             assign("est", est, .env);
             assign("stopTime", Sys.time(), .env);
+            assign("origControl", control, .env);
+            assign("modelId", .modelId, .env);
         }
         return(.ret);
     } else if (est == "nlme" || est == "nlme.mu" || est == "nlme.mu.cov" || est == "nlme.free"){
+        if (.nTv != 0) stop("nlme does not support time-varying covariates (yet)");
+        data <- as.data.frame(data)
+        if (length(uif$predDf$cond) > 1) stop("nlmixr nlme does not support multiple endpoints.")
         pt <- proc.time()
         est.type <- est;
         if (est == "nlme.free"){
@@ -481,36 +522,32 @@ nlmixr_fit <- function(uif, data, est=NULL, control=list(), ...,
         }));
         dat$nlmixr.num <- seq_along(dat$nlmixr.grp)
         weight <- uif$nlme.var
-        if (!is.null(uif$nmodel$lin.solved)){
-            fit <- nlme_lin_cmpt(dat, par_model=specs,
-                                 ncmt=uif$nmodel$lin.solved$ncmt,
-                                 oral=uif$nmodel$lin.solved$oral,
-                                 tlag=uif$nmodel$lin.solved$tlag,
-                                 infusion=uif$env$infusion,
-                                 parameterization=uif$nmodel$lin.solved$parameterization,
-                                 par_trans=fun,
-                                 weight=weight,
-                                 verbose=TRUE,
-                                 control=control,
-                                 ...);
+        if (sum.prod){
+            rxode <- RxODE::rxSumProdModel(uif$rxode.pred);
         } else {
-            if (sum.prod){
-                rxode <- RxODE::rxSumProdModel(uif$rxode.pred);
-            } else {
-                rxode <- uif$rxode.pred;
-            }
-            fit <- nlme_ode(dat,
-                            model=rxode,
-                            par_model=specs,
-                            par_trans=fun,
-                            response="nlmixr_pred",
-                            weight=weight,
-                            verbose=TRUE,
-                            control=control,
-                            ...);
+            rxode <- uif$rxode.pred;
         }
+        .atol <- 1e-8
+        if (!is.null(control$atol)) .atol <- control$atol
+        .rtol <- 1e-8
+        if (!is.null(control$rtol)) .rtol <- control$rtol
+        fit <- nlme_ode(dat,
+                        model=rxode,
+                        par_model=specs,
+                        par_trans=fun,
+                        response="nlmixr_pred",
+                        weight=weight,
+                        verbose=TRUE,
+                        control=control,
+                        atol=.atol,
+                        rtol=.rtol,
+                        ...);
         class(fit) <- c(est.type, class(fit));
-        .ret <- as.focei.nlmixrNlme(fit, uif, pt, data=dat, calcResid=calc.resid);
+        .ret <- try({as.focei.nlmixrNlme(fit, uif, pt, data=dat, calcResid=calc.resid)});
+        if (inherits(.ret, "try-error")){
+            warning("Error converting to nlmixr UI object, returning nlme object");
+            return(fit);
+        }
         if (inherits(.ret, "nlmixrFitData")){
             .ret <- fix.dat(.ret);
             .ret <- .addNpde(.ret);
@@ -520,6 +557,8 @@ nlmixr_fit <- function(uif, data, est=NULL, control=list(), ...,
             assign("startTime", start.time, .env);
             assign("est", est, .env);
             assign("stopTime", Sys.time(), .env);
+            assign("origControl",control,.env);
+            assign("modelId",.modelId,.env);
         }
         return(.ret)
     } else if (any(est == c("foce", "focei", "fo", "foi"))){
@@ -554,7 +593,7 @@ nlmixr_fit <- function(uif, data, est=NULL, control=list(), ...,
             .default <- foceiControl();
             control$maxInnerIterations <- .default$maxInnerIterations
             control$maxOuterIterations <- 0L;
-            control$covMethod <- "";
+            control$covMethod <- 0L;
             control$fo <- 0L
             .uif <- fit$uif;
             .thetas <- fit$theta;
@@ -569,12 +608,18 @@ nlmixr_fit <- function(uif, data, est=NULL, control=list(), ...,
             }
             .time <- fit$time;
             .objDf <- fit$objDf;
+            .message <- fit$env$message
+            .time <- fit$time
             env <- new.env(parent=emptyenv());
-            env$cov <- fit$cov;
+            for (.w in c("cov", "covR", "covS", "covMethod")){
+                if (exists(.w, fit$env)){
+                    assign(.w, get(.w, envir=fit$env), envir=env);
+                }
+            }
             env$time2 <- time;
             env$uif <- .uif;
             env$method <- "FO";
-            fit <- foceiFit(dat,
+            fit0<- try(foceiFit(dat,
                             inits=.uif$focei.inits,
                             PKpars=.uif$theta.pars,
                             ## par_trans=fun,
@@ -588,23 +633,38 @@ nlmixr_fit <- function(uif, data, est=NULL, control=list(), ...,
                             etaNames=.uif$eta.names,
                             control=control,
                             env=env,
-                            ...);
-            .tmp1 <- env$objDf;
-            if (any(names(.objDf) == "Condition Number")) .tmp1 <- data.frame(.tmp1, "Condition Number"=NA, check.names=FALSE);
-            if (any(names(.tmp1) == "Condition Number")) .objDf <- data.frame(.objDf, "Condition Number"=NA, check.names=FALSE);
-            env$objDf <- rbind(.tmp1, .objDf);
-            row.names(env$objDf) <- c(ifelse(est == "fo", "FOCE", "FOCEi"), "FO");
+                            ...), silent = TRUE);
+            if (inherits(fit0, "try-error")){
+            } else {
+                fit <- fit0
+                assign("message2", fit$env$message, env);
+                assign("message", .message, env);
+                .tmp1 <- env$objDf;
+                if (any(names(.objDf) == "Condition Number")) .tmp1 <- data.frame(.tmp1, "Condition Number"=NA, check.names=FALSE);
+                if (any(names(.tmp1) == "Condition Number")) .objDf <- data.frame(.objDf, "Condition Number"=NA, check.names=FALSE);
+                env$objDf <- rbind(.tmp1, .objDf);
+                row.names(env$objDf) <- c(ifelse(est == "fo", "FOCE", "FOCEi"), "FO");
+                .tmp1 <- env$time;
+                .tmp1$optimize <- .time$optimize
+                .tmp1$covariance <- .time$covariance
+                assign("time", .tmp1, envir=env)
+                env$objDf <- env$objDf[, c("OBJF", "AIC", "BIC", "Log-likelihood", "Condition Number")]
+                setOfv(env, "fo")
+            }
         }
-        fit <- fix.dat(fit);
         fit <- .addNpde(fit);
         assign("start.time", start.time, env);
         assign("est", est, env);
         assign("stop.time", Sys.time(), env);
+        assign("origControl",control, env);
+        assign("modelId",.modelId, env);
         return(fit);
     } else if (est == "posthoc"){
-        control <- do.call(nlmixr::foceiControl, control);
+        if (class(control) !="foceiControl") control <- do.call(nlmixr::foceiControl, control);
         control$covMethod <- 0L
         control$maxOuterIterations <- 0L;
+        .env <- new.env(parent=emptyenv());
+        .env$uif <- uif;
         fit <- foceiFit(dat,
                         inits=uif$focei.inits,
                         PKpars=uif$theta.pars,
@@ -617,16 +677,106 @@ nlmixr_fit <- function(uif, data, est=NULL, control=list(), ...,
                         thetaNames=uif$focei.names,
                         etaNames=uif$eta.names,
                         control=control,
+                        env=.env,
                         ...)
+        if (inherits(fit, "nlmixrFitData")){
+            .cls <- class(fit);
+            .env <- attr(.cls, ".foceiEnv");
+            .cls[1]  <- "nlmixrPosthoc"
+            class(fit) <- .cls
+        }
+        assign("uif", uif, fit$env)
         ## assign("start.time", start.time, env);
         ## assign("est", est, env);
         ## assign("stop.time", Sys.time(), env);
         ## assign("start.time", start.time, env);
         ## assign("est", est, env);
         ## assign("stop.time", Sys.time(), env);
+        assign("origControl",control,fit$env);
+        assign("modelId",.modelId,fit$env);
         return(fit);
+    } else {
+        stop(sprintf("Unknown estimation method est=\"%s\"",est));
     }
 }
+
+##' Fit a nlmixr model
+##'
+##' @param data Dataset to estimate.  Needs to be RxODE compatible in
+##'     EVIDs.
+##' @param uif Parsed nlmixr model (by \code{nlmixr(mod.fn)}).
+##' @param est Estimation method
+##' @param control Estimation control options.  They could be
+##'     \code{\link[nlme]{nlmeControl}}, \code{\link{saemControl}} or
+##'     \code{\link{foceiControl}}
+##' @param ... Parameters passed to estimation method.
+##' @param sum.prod Take the RxODE model and use more precise
+##'     products/sums.  Increases solving accuracy and solving time.
+##' @param table A list controlling the table options (i.e. CWRES,
+##'     NPDE etc).  See \code{\link{tableControl}}.
+##' @param save This option determines if the fit will be saved to be
+##'     reloaded if already run.  If NULL, get the option from
+##'     \code{options("nlmixr.save")};
+##' @return nlmixr fit object
+##' @author Matthew L. Fidler
+##' @export
+nlmixr_fit  <- function(uif, data, est=NULL, control=list(), ...,
+                        sum.prod=FALSE, table=tableControl(),
+                        save=NULL){
+
+    if (is.null(save)){
+        save <- getOption("nlmixr.save", FALSE);
+    }
+    if (save){
+        .modName  <- ifelse(is.null(uif$model.name),"",paste0(uif$model.name,"-"));
+        if (.modName==".-") .modName <- ""
+        .dataName  <- ifelse(is.null(uif$data.name),"",paste0(uif$data.name,"-"));
+        if (.dataName==".-") .dataName <- ""
+        .digest <- digest::digest(list(gsub("<-","=",gsub(" +","",uif$fun.txt)),
+                                       as.data.frame(uif$ini),
+                                       data,
+                                       est,
+                                       control,
+                                       sum.prod,
+                                       table,
+                                       ...,
+                                       as.character(utils::packageVersion("nlmixr")),
+                                       as.character(utils::packageVersion("RxODE"))))
+        .saveFile  <- file.path(getOption("nlmixr.save.dir", getwd()),
+                                paste0("nlmixr-",.modName,.dataName,est,"-",.digest,".rds"));
+        if (file.exists(.saveFile)){
+            message(sprintf("Loading model already run (%s)",.saveFile))
+            .ret  <- readRDS(.saveFile)
+            return(.ret)
+        }
+    }
+    .ret  <- .collectWarnings(nlmixr_fit0(uif=uif, data=data, est=est, control=control, ...,
+                                          sum.prod=sum.prod, table=table), TRUE);
+    .ws  <- .ret[[2]];
+    .ret  <- .ret[[1]];
+    if (inherits(.ret, "nlmixrFitCore")){
+        .env <- .ret$env
+        .env$warnings <- .ws
+    }
+    for (.i in seq_along(.ws)){
+        warning(.ws[.i])
+    }
+    if (inherits(.ret, "nlmixrFitCore")){
+        if (save){
+            AIC(.ret); # Calculate SAEM AIC when saving...
+            .env <- .ret$env
+            .extra <- (proc.time() - .nlmixrTime)["elapsed"] - sum(.env$time)
+            .env$time <- data.frame(.env$time,"other"=.extra, check.names=FALSE)
+            saveRDS(.ret,file=.saveFile)
+        } else {
+            .env <- .ret$env
+            .extra <- (proc.time() - .nlmixrTime)["elapsed"] - sum(.env$time)
+            .env$time <- data.frame(.env$time,"other"=.extra, check.names=FALSE)
+        }
+    }
+    return(.ret);
+}
+
 ##' Control Options for SAEM
 ##'
 ##' @param seed Random Seed for SAEM step.  (Needs to be set for
@@ -667,6 +817,8 @@ nlmixr_fit <- function(uif, data, est=NULL, control=list(), ...,
 ##'     gradient cross-product (evaluated at the individual empirical
 ##'     Bayes estimates).
 ##'
+##'  "\code{linFim}" Use the Linearized Fisher Information Matrix to calculate the covariance.
+##'
 ##'  "\code{fim}" Use the SAEM-calculated Fisher Information Matrix to calculate the covariance.
 ##'
 ##'  "\code{r,s}" Uses the sandwich matrix to calculate the covariance, that is: \eqn{R^-1 \times S \times R^-1}
@@ -676,14 +828,31 @@ nlmixr_fit <- function(uif, data, est=NULL, control=list(), ...,
 ##'  "\code{s}" Uses the crossproduct matrix to calculate the covariance as \eqn{4\times S^-1}
 ##'
 ##'  "" Does not calculate the covariance step.
+##'
 ##' @param logLik boolean indicating that log-likelihood should be
 ##'     calculate by Gaussian quadrature.
+##'
 ##' @param trace An integer indicating if you want to trace(1) the
 ##'     SAEM algorithm process.  Useful for debugging, but not for
 ##'     typical fitting.
+##'
+##' @param nnodes.gq number of nodes to use for the Gaussian
+##'     quadrature when computing the likelihood with this method
+##'     (defaults to 1, equivalent to the Laplaclian likelihood)
+##'
+##' @param nsd.gq span (in SD) over which to integrate when computing
+##'     the likelihood by Gaussian quadrature. Defaults to 3 (eg 3
+##'     times the SD)
+##'
+##' @param adjObf is a boolean to indicate if the objective function
+##'     should be adjusted to be closer to NONMEM's default objective
+##'     function.  By default this is \code{TRUE}
+##'
 ##' @param ... Other arguments to control SAEM.
+##'
 ##' @inheritParams RxODE::rxSolve
 ##' @inheritParams foceiControl
+##'
 ##' @return List of options to be used in \code{\link{nlmixr}} fit for
 ##'     SAEM.
 ##' @author Wenping Wang & Matthew L. Fidler
@@ -694,13 +863,17 @@ saemControl <- function(seed=99,
                         nu=c(2, 2, 2),
                         atol = 1e-06,
                         rtol = 1e-04,
-                        stiff = TRUE,
+                        method = "lsoda",
                         transitAbs = FALSE,
                         print=1,
                         trace=0,
-                        covMethod=c("fim", "r,s", "r", "s"),
+                        covMethod=c("linFim", "fim", "r,s", "r", "s"),
                         logLik=FALSE,
+                        nnodes.gq=3,
+                        nsd.gq=1.6,
                         optExpression=TRUE,
+                        maxsteps=100000L,
+                        adjObf=TRUE,
                         ...){
     .xtra <- list(...);
     .rm <- c();
@@ -717,13 +890,16 @@ saemControl <- function(seed=99,
         .rm <- c(.rm, "n.em")
     }
     .ret <- list(mcmc=list(niter=c(nBurn, nEm), nmc=nmc, nu=nu),
-         ODEopt=list(atol=atol, rtol=rtol,
-                     stiff=as.integer(stiff),
-                     transitAbs = as.integer(transitAbs)),
-         seed=seed,
-         print=print,
-         DEBUG=trace,
-         optExpression=optExpression, ...)
+                 ODEopt=RxODE::rxControl(atol=atol, rtol=rtol, method=method,
+                                         transitAbs = transitAbs,maxsteps=maxsteps,...),
+                 seed=seed,
+                 print=print,
+                 DEBUG=trace,
+                 optExpression=optExpression,
+                 nnodes.gq=nnodes.gq,
+                 nsd.gq=nsd.gq,
+                 adjObf=adjObf,
+                 ...)
     if (length(.rm) > 0){
         .ret <- .ret[!(names(.ret) %in% .rm)]
     }
@@ -738,34 +914,64 @@ saemControl <- function(seed=99,
 ##' This returns a new fit object with CWRES attached
 ##'
 ##' @param fit nlmixr fit without WRES/CWRES
-##' @param updateObject Boolean indicating if the original fit object should be updated. By default this is true.
+##' @param updateObject Boolean indicating if the original fit object
+##'     should be updated. By default this is true.
+##' @param envir Environment that should be checked for object to
+##'     update.  By default this is the global environment.
 ##' @return fit with CWRES
 ##' @author Matthew L. Fidler
 ##' @export
-addCwres <- function(fit, updateObject=TRUE){
+addCwres <- function(fit, updateObject=TRUE, envir=globalenv()){
+    .pt  <- proc.time();
+    .oTime <- fit$env$time;
     .objName <- substitute(fit);
     if(any(names(fit) == "CWRES")){
-        warning("Already contains CWRES");
+        ## warning("Already contains CWRES");
         return(fit)
     }
     .uif <- fit$uif;
     .saem <- fit$saem
+    .od <- fit$origData
     if (!is.null(.saem)){
-        .newFit <- as.focei.saemFit(.saem, .uif, data=getData(fit), calcResid = TRUE, obf=fit$objDf["SAEMg","OBJF"]);
+        assign("saem",NULL,fit$env)
+        on.exit({assign("saem",.saem,fit$env)});
+        .newFit <- as.focei.saemFit(.saem, .uif, data=.nmGetData(fit), calcResid = TRUE, obf=NA, calcCov=FALSE);
+        .ob1 <- .newFit$objDf
+        .ob2 <- fit$objDf
+        if (any(names(.ob2) == "Condition Number")){
+            .cn <- unique(.ob2[["Condition Number"]])
+            .cn <- .cn[!is.na(.cn)];
+            if (length(.cn)==1){
+                .ob1[,"Condition Number"] <- .cn
+            } else {
+                .ob1[,"Condition Number"] <- NA;
+            }
+        }
+        .ob1 <- rbind(.ob1,.ob2);
+        .ob1 <- .ob1[order(row.names(.ob1)),];
+        .ob1 <- .ob1[!is.na(.ob1$OBJF),];
+        assign("objDf", .ob1, envir=.newFit$env);
+        assign("saem",.saem,fit$env);
+        assign("adjObj",fit$env$adjObj, .newFit$env)
         .df <- .newFit[, c("WRES", "CRES", "CWRES", "CPRED")];
+        ## Add CWRES timing to fit.
         .new <- cbind(fit, .df);
     }
     .nlme <- fit$nlme
     if (!is.null(.nlme)){
-        .newFit <- as.focei(.nlme, .uif, data=getData(fit), calcResid = TRUE);
+        .newFit <- as.focei(.nlme, .uif, data=.nmGetData(fit), calcResid = TRUE);
+        assign("adjObj",fit$env$adjObj, .newFit$env)
         .df <- .newFit[, c("WRES", "CRES", "CWRES", "CPRED")];
         .new <- cbind(fit, .df);
     }
     class(.new) <- class(.newFit)
+    if (!is.null(.saem)){
+        setOfv(.new, "FOCEi");
+    }
     if (updateObject){
-        .parent <- parent.frame(2);
+        .parent <- envir;
         .bound <- do.call("c", lapply(ls(.parent, all.names=TRUE), function(.cur){
-                                   if (.cur == .objName && identical(.parent[[.cur]], fit)){
+                                   if (.cur == .objName && identical(.parent[[.cur]]$env, fit$env)){
                                        return(.cur)
                                    }
                                    return(NULL);
@@ -776,5 +982,8 @@ addCwres <- function(fit, updateObject=TRUE){
             }
         }
     }
+    .env <- .new$env
+    .env$time <- data.frame(.oTime,cwres=(proc.time() - .pt)["elapsed"],check.names=FALSE);
+    assign("origData", .od, .env)
     return(.new);
 }
